#include <getopt.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

typedef struct disc_file_descriptor_t disc_file_descriptor_t;

#include "afs.h"
#include "memory.h"

static const char* opt_file = 0;
static const char* opt_chdir = 0;
static int opt_verbose = 0;
static char opt_mode = 0;
static int opt_strip_components = 0;

static afs_volume_t vol;

void AdtMessageBox(const char* fmt, ...) {
	va_list args;

    va_start(args, fmt);
    vprintf(fmt, args);
    va_end(args);
}

disc_file_descriptor_t* DiscOpenFile(const char* path) {
	return (disc_file_descriptor_t*) fopen(path, "r");
}

int DiscReadFile(disc_file_descriptor_t* handle, void* buffer, int length) {
	fread(buffer, length, 1, (FILE*) handle);
	return 0;
}

int DiscSeekFile(disc_file_descriptor_t* handle, int offset, int whence) {
	int origin;
	switch (whence) {
		case 0:
			origin = SEEK_SET;
			break;
		case 1:
			origin = SEEK_CUR;
			break;
		case 2:
			origin = SEEK_END;
			break;
		default:
			abort();
	}

	return fseek((FILE*) handle, offset, origin);
}

void* MemoryAllocate(int size) {
	return malloc(size);
}

void MemoryFree(void* buffer) {
	free(buffer);
}

void walk_afs_volume(afs_volume_t* vol, void(*callback)(afs_volume_t* vol, char* name, int isfile)) {
	int depth = 0;
	char path[128] = { 0 };

	strcpy(path, vol->entries[0].name);
	vol->entries[0].field1_0xc = 0;
	callback(vol, path, 0);
	strcat(path, "\\");

	for (int index = 1; index < vol->numberEntries; index++) {
		afs_entry_t* entry = &vol->entries[index];
		afs_entry_t* parent = &vol->entries[atoi(entry->name + 1)];
		char* component = strchr(entry->name, '_') + 1;

		char* target = path;
		for (int count = parent->field1_0xc; count >= 0; count--) {
			target = strstr(target, "\\") + strlen("\\");
		}

		strcpy(target, component);

		if (entry->location == 0xCDCDCDCD) {
			entry->field1_0xc = parent->field1_0xc + 1;
			callback(vol, path, 0);
			strcat(path, "\\");
		}
		else {
			callback(vol, path, 1);
		}
	}
}

void my_callback(afs_volume_t* vol, char* name, int isfile) {
	char systemname[80];
	strcpy(systemname, name);
	char *partname = systemname;
	char *sep;

	for (int i = opt_strip_components; i > 0; i--) {
		partname = strstr(partname, "\\");
		if (partname == 0) {
			return;
		}
		partname++;
	}

	if (opt_mode == 't' || opt_verbose >= 1) {
		puts(partname);
	}

#if !defined(WIN32) && !defined(_WIN32)
	while ((sep = strstr(systemname, "\\"))) {
		*sep = '/';
	}
#endif

	if (opt_mode == 'x') {
		if (isfile == 0) {
			if (mkdir(partname, 0755) != 0) {
				perror(partname);
			}
		}
		else {
			afs_file_descriptor_t* afd = AfsOpen(vol, name);
			int length = AfsFileSize(vol, afd);
			void* buffer = MemoryAllocate(length);
			AfsRead(vol, afd, buffer, length);
			AfsClose(afd);

			int fd = open(partname, O_WRONLY|O_CREAT|O_TRUNC, 0755);
			if (fd == -1) {
				perror(partname);
			}
			write(fd, buffer, length);
			close(fd);

			MemoryFree(buffer);
		}
	}
}

int main(int argc, char* argv[]) {
	int c;

	while ((c = getopt(argc, argv, "C:f:s:tvx")) != -1) {
		switch (c) {
			case 'C':
				opt_chdir = optarg;
				break;
			case 'f':
				opt_file = optarg;
				break;
			case 's':
				opt_strip_components = atoi(optarg);
				break;
			case 'v':
				opt_verbose = 1;
				break;
			case 't':
			case 'x':
				opt_mode = c;
				break;
			default:
				abort();
		}
	}

	if (opt_file == 0 || opt_mode == 0) {
		exit(EXIT_FAILURE);
	}

	if (AfsOpenVolume(&vol, opt_file) == -1) {
		exit(EXIT_FAILURE);
	}

	if (opt_chdir != 0) {
		if (chdir(opt_chdir) != 0) {
			perror(opt_chdir);
			exit(EXIT_FAILURE);
		}
	}

	walk_afs_volume(&vol, my_callback);

	return EXIT_SUCCESS;
}
